/*
 *  Copyright (C) 2014 Andreas Steinmetz <ast@domdv.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <X11/Xlib.h>
#include <X11/XKBlib.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <time.h>

static char *mtab[]=
{
	"shift",
	"lock",
	"control",
	"mod1",
	"mod2",
	"mod3",
	"mod4",
	"mod5",
	NULL
};

static char *display=NULL;
static char *key=NULL;
static char *mod=NULL;

static void worker(union sigval unused)
{
	int i;
	Display *dpy;
	KeySym s;
	KeyCode k;
	XModifierKeymap *m;

	for(i=0;mtab[i];i++)if(!strcasecmp(mod,mtab[i]))break;
	if(!mtab[i])return;
	if((s=XStringToKeysym(key))==NoSymbol)return;
	if(!(dpy=XOpenDisplay(display)))return;
	if(!(k=XKeysymToKeycode(dpy,s)))goto out;
	if(!(m=XGetModifierMapping(dpy)))goto out;
	if(!XDeleteModifiermapEntry(m,k,i))goto out2;
	XSetModifierMapping(dpy,m);
out2:	XFreeModifiermap(m);
out:	XCloseDisplay(dpy);
}

static void usage(char *cmd)
{
	fprintf(stderr,"Usage: %s [-f] [-d display] -k key -m modifier\n",cmd);
	exit(1);
}

int main(int argc,char *argv[])
{
	int c;
	int fg=0;
	int major=XkbMajorVersion;
	int minor=XkbMinorVersion;
	int err;
	Display *dpy;
	XkbEvent ev;
	struct sigevent sev;
	timer_t tid;
	struct itimerspec its;
	union sigval unused;

	while((c=getopt(argc,argv,"k:m:d:f"))!=-1)switch(c)
	{
	case 'k':
		key=optarg;
		break;
	case 'm':
		mod=optarg;
		break;
	case 'd':
		display=optarg;
		break;
	case 'f':
		fg=1;
		break;
	default:usage(argv[0]);
	}

	if(!key||!mod)usage(argv[0]);

	if(!fg)if(daemon(0,0))
	{
		perror("daemon");
		return 1;
	}

	if(!(dpy=XkbOpenDisplay(display,NULL,&err,&major,&minor,NULL)))
	{
		fprintf(stderr,"XkbOpenDisplay failed with code %d\n",err);
		return 1;
	}

	if(!XkbSelectEvents(dpy,XkbUseCoreKbd,XkbNewKeyboardNotifyMask,
		XkbNewKeyboardNotifyMask))
	{
		fprintf(stderr,"XkbSelectEvents failed\n");
		return 1;
	}

	memset(&sev,0,sizeof(sev));
	sev.sigev_notify=SIGEV_THREAD;
	sev.sigev_notify_function=worker;

	if(timer_create(CLOCK_MONOTONIC,&sev,&tid))
	{
		perror("timer_create");
		return 1;
	}

	memset(&its,0,sizeof(its));

	worker(unused);

	while(!XNextEvent(dpy,&ev.core))
	{
		its.it_value.tv_nsec=0;
		timer_settime(tid,0,&its,NULL);
		its.it_value.tv_nsec=100000000;
		timer_settime(tid,0,&its,NULL);
	}

	XCloseDisplay(dpy);

	return 0;
}
